/*
 * SIP Communicator, the OpenSource Java VoIP and Instant Messaging client.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package net.java.sip.communicator.plugin.customjabberwizz;

import java.util.*;

import net.java.sip.communicator.plugin.jabberaccregwizz.*;
import net.java.sip.communicator.service.gui.*;
import net.java.sip.communicator.service.ldap.LdapConstants.Auth;
import net.java.sip.communicator.service.ldap.LdapConstants.Encryption;
import net.java.sip.communicator.service.ldap.LdapConstants.Scope;
import net.java.sip.communicator.service.ldap.*;
import net.java.sip.communicator.service.protocol.*;

/**
 * The <tt>CustomJabberAccountRegistrationWizard</tt> is an implementation of
 * the <tt>AccountRegistrationWizard</tt> for Jabber based custom network. It
 * should allow the user to create and configure a new custom Jabber account.
 *
 * @author Yana Stamcheva
 */
public class CustomJabberAccountRegistrationWizard
    extends JabberAccountRegistrationWizard
{
    /**
     * The Google Talk protocol name.
     */
    public static final String PROTOCOL = CustomJabberAccRegWizzActivator
        .getResources().getI18NString("plugin.customjabberwizz.PROTOCOL_NAME");

    /**
     * A constant pointing to the Google Talk protocol logo image.
     */
    public static final String PROTOCOL_ICON
        = "plugin.customjabberwizz.ACCOUNT_ICON_16x16";

    /**
     * A constant pointing to the Aim protocol wizard page image.
     */
    public static final String PAGE_IMAGE
        = "plugin.customjabberwizz.ACCOUNT_ICON_64x64";

    private JabberAccountCreationForm createAccountService;

    /**
     * Creates an instance of <tt>CustomJabberAccountRegistrationWizard</tt>.
     * @param wizardContainer the wizard container, where this wizard
     * is added
     */
    public CustomJabberAccountRegistrationWizard(
        WizardContainer wizardContainer)
    {
        super(wizardContainer);
    }

    /**
     * Returns the set of pages contained in this wizard.
     * @return Iterator
     */
    public Iterator<WizardPage> getPages()
    {
        JabberAccountRegistration reg = new JabberAccountRegistration();

        setPredefinedProperties(reg);

        return getPages(reg);
    }

    /**
     * Returns the first wizard page.
     *
     * @return the first wizard page.
     */
    public Object getSimpleForm(boolean isCreateAccount)
    {
        JabberAccountRegistration reg = new JabberAccountRegistration();

        setPredefinedProperties(reg);

        return getSimpleForm(reg, isCreateAccount);
    }

    /**
     * Sets all google talk specific properties.
     *
     * @param reg the registration object
     */
    private void setPredefinedProperties(JabberAccountRegistration reg)
    {
        reg.setDefaultUserSufix(CustomJabberAccRegWizzActivator.getResources()
            .getSettingsString("plugin.customjabberwizz.DEFAULT_USER_SUFFIX"));
        reg.setServerOverridden(true);

        String sufix = CustomJabberAccRegWizzActivator.getResources()
            .getSettingsString("plugin.customjabberwizz.OVERRIDE_PHONE_SUFFIX");

        if(sufix != null && sufix.length() > 0)
        {
            reg.setOverridePhoneSufix(sufix);
        }

        String bypassCaps
            = CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.TELEPHONY_BYPASS_GTALK_CAPS");

        if(bypassCaps != null && bypassCaps.length() > 0)
        {
            reg.setTelephonyDomainBypassCaps(bypassCaps);
        }

        String priority
            = CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.OVERRIDE_PRIORITY");
        if(priority != null && priority.length() > 0)
        {
            reg.setPriority(Integer.parseInt(priority));
        }

        String smsServerAddressKey
            = "net.java.sip.communicator.plugin.customjabberwizz.SMS_SERVER_ADDRESS";

        // Try to obtain the SMS server address from the configuration service.
        String smsServerAddress
            = CustomJabberAccRegWizzActivator.getConfigService()
                            .getString(smsServerAddressKey);

        // If it's not yet set in the configuration service, we'll have
        // a look in the default properties.
        if (smsServerAddress == null || smsServerAddress.length() <= 0)
        {
            smsServerAddress
                = CustomJabberAccRegWizzActivator.getResources()
                    .getSettingsString(
                        "plugin.customjabberwizz.SMS_SERVER_ADDRESS");
        }

        if(smsServerAddress != null && smsServerAddress.length() > 0)
        {
            reg.setSmsServerAddress(smsServerAddress);
        }
    }

    /**
     * Installs the account defined in this wizard.
     *
     * @param userName the user name to sign in with
     * @param password the password to sign in with
     * @return the created <tt>ProtocolProviderService</tt> corresponding to the
     * new account
     * @throws OperationFailedException if the operation didn't succeed
     */
    public ProtocolProviderService signin(String userName, String password)
        throws OperationFailedException
    {
        setPredefinedProperties(getRegistration());

        return super.signin(userName, password);
    }

    /**
     * Implements the <code>AccountRegistrationWizard.getIcon</code> method.
     * Returns the icon to be used for this wizard.
     * @return byte[]
     */
    public byte[] getIcon()
    {
        return CustomJabberAccRegWizzActivator.getResources()
            .getImageInBytes(PROTOCOL_ICON);
    }
    
    /**
     * Implements the <code>AccountRegistrationWizard.getPageImage</code> method.
     * Returns the image used to decorate the wizard page
     * 
     * @return byte[] the image used to decorate the wizard page
     */
    public byte[] getPageImage()
    {
        return CustomJabberAccRegWizzActivator.getResources()
                .getImageInBytes(PAGE_IMAGE);
    }

    /**
     * Implements the <code>AccountRegistrationWizard.getProtocolName</code>
     * method. Returns the protocol name for this wizard.
     * @return String
     */
    public String getProtocolName()
    {
        return CustomJabberAccRegWizzActivator.getResources()
                .getI18NString("plugin.customjabberwizz.PROTOCOL_NAME");
    }

    /**
     * Implements the <code>AccountRegistrationWizard.getProtocolDescription
     * </code> method. Returns the description of the protocol for this wizard.
     * @return String
     */
    public String getProtocolDescription()
    {
        return CustomJabberAccRegWizzActivator.getResources()
            .getI18NString("plugin.customjabberwizz.PROTOCOL_DESCRIPTION");
    }

    /**
     * Returns an example string, which should indicate to the user how the
     * user name should look like.
     * @return an example string, which should indicate to the user how the
     * user name should look like.
     */
    public String getUserNameExample()
    {
        return CustomJabberAccRegWizzActivator.getResources()
            .getI18NString("plugin.customjabberwizz.USERNAME_EXAMPLE");
    }

    /**
     * Returns the display label used for the sip id field.
     * @return the sip id display label string.
     */
    protected String getUsernameLabel()
    {
        return CustomJabberAccRegWizzActivator.getResources()
            .getI18NString("plugin.customjabberwizz.USERNAME");
    }

    /**
     * Returns the protocol name as listed in "ProtocolNames" or just the name
     * of the service.
     * @return the protocol name
     */
    public String getProtocol()
    {
        return PROTOCOL;
    }

    /**
     * Returns the protocol icon path.
     * @return the protocol icon path
     */
    public String getProtocolIconPath()
    {
        return CustomJabberAccRegWizzActivator.getResources().getI18NString(
            "plugin.customjabberwizz.PROTOCOL_ICON_PATH");
    }

    /**
     * Returns the account icon path.
     * @return the account icon path
     */
    public String getAccountIconPath()
    {
        return  CustomJabberAccRegWizzActivator.getResources().getI18NString(
            "plugin.customjabberwizz.ACCOUNT_ICON_PATH");
    }

    /**
     * Opens a browser on the sign up page. 
     */
    public void webSignup()
    {
        CustomJabberAccRegWizzActivator.getBrowserLauncher()
            .openURL(CustomJabberAccRegWizzActivator.getResources()
                .getI18NString("plugin.customjabberwizz.CREATE_ACCOUNT_LINK"));
    }

    /**
     * Returns <code>true</code> if the web sign up is supported by the current
     * implementation, <code>false</code> - otherwise.
     * @return <code>true</code> if the web sign up is supported by the current
     * implementation, <code>false</code> - otherwise
     */
    public boolean isWebSignupSupported()
    {
        return Boolean.valueOf(
            CustomJabberAccRegWizzActivator.getResources().getI18NString(
                "plugin.customjabberwizz.IS_WEB_SIGN_UP")).booleanValue();
    }

    /**
     * Return the string for home page link label.
     * @return the string for home page link label
     */
    protected String getHomeLinkLabel()
    {
        return  CustomJabberAccRegWizzActivator.getResources()
                    .getI18NString("plugin.customjabberwizz.HOME_LINK");
    }

    /**
     * Return the string for create new account button.
     * @return the string for create new account button.
     */
    protected String getCreateAccountLabel()
    {
        return CustomJabberAccRegWizzActivator.getResources().getI18NString(
            "plugin.customjabberwizz.REGISTER_NEW_ACCOUNT_TEXT");
    }

    /**
     * Return the string for add existing account button.
     * @return the string for add existing account button.
     */
    protected String getCreateAccountButtonLabel()
    {
        return CustomJabberAccRegWizzActivator.getResources().getI18NString(
            "plugin.customjabberwizz.NEW_ACCOUNT_TITLE");
    }

    /**
     * Returns an instance of <tt>CreateAccountService</tt> through which the
     * user could create an account. This method is meant to be implemented by
     * specific protocol provider wizards.
     * @return an instance of <tt>CreateAccountService</tt>
     */
    protected JabberAccountCreationFormService getCreateAccountService()
    {
        String isCreateAccFormProp
            = CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.CREATE_ACCOUNT_SERVICE");

        boolean isCreateAccForm = false;
        if (isCreateAccFormProp != null && isCreateAccFormProp.length() > 0)
        {
            isCreateAccForm = Boolean.parseBoolean(isCreateAccFormProp);
        }

        if (!isCreateAccForm)
            return null;

        if (createAccountService == null)
            createAccountService = new JabberAccountCreationForm();

        return createAccountService;
    }

    /**
     * Return the string for add existing account button.
     * @return the string for add existing account button.
     */
    protected String getExistingAccountLabel()
    {
        return Resources.getString("plugin.customjabberwizz.EXISTING_ACCOUNT");
    }

    /**
     * Indicates that the account corresponding to the given
     * <tt>protocolProvider</tt> has been removed.
     * @param protocolProvider the protocol provider that has been removed
     */
    public void accountRemoved(ProtocolProviderService protocolProvider)
    {
        if (CustomJabberAccRegWizzActivator.isCustomJabberAccountAvailable())
            return;

        // If there's no more accounts from this protocol and an ldap server
        // exists for this protocol we remove it.
        LdapDirectorySet serverSet =
            CustomJabberAccRegWizzActivator.getLdapService().getServerSet();

        String ldapName
            = protocolProvider.getAccountID().getAccountAddress() + " ("
                + CustomJabberAccRegWizzActivator.getResources()
                    .getSettingsString("plugin.customjabberwizz.LDAP_NAME")
                + ")";

        LdapDirectory server
            = serverSet.getServerWithName(ldapName);

        // If the ldap server exists, we'll re-init it.
        if (server != null)
        {
            serverSet.removeServerWithName(server.getSettings().getName());
            new RefreshContactSourceThread(server, null).start();
        }
    }

    /**
     * Creates an account for the given user and password.
     *
     * @param providerFactory the ProtocolProviderFactory which will create
     * the account
     * @param userName the user identifier
     * @param passwd the password
     * @return the <tt>ProtocolProviderService</tt> for the new account.
     * @throws OperationFailedException if the operation didn't succeed
     */
    protected ProtocolProviderService installAccount(
        ProtocolProviderFactory providerFactory,
        String userName,
        String passwd)
        throws OperationFailedException
    {
        JabberAccountRegistration registration = getRegistration();

        // Check if a default STUN/TURN server account should be created
        // together with this account and if yes, create the STUN server given
        // by the default setting parameters.
        String createStunAccString
            = CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.CREATE_STUN_TURN_ACCOUNT");

        if(createStunAccString != null && createStunAccString.length() > 0)
        {
            boolean isCreateStunAcc
                = new Boolean(createStunAccString).booleanValue();

            if (isCreateStunAcc)
                addDefaultStunServerAccount(registration);
        }

        setPredefinedProperties(registration);

        String enableShortDisplayName
            = CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.ENABLE_SHORT_DISPLAY_NAME");

        if(enableShortDisplayName != null && enableShortDisplayName.length() > 0)
        {
            if (Boolean.parseBoolean(enableShortDisplayName))
                getRegistration().setAccountDisplayName(
                    (userName.indexOf('@') < 0) ? userName
                    : userName.substring(0, userName.indexOf('@'))
                    + " (" + getProtocolName() + ")");
        }

        // Call the parent install account that would set all account
        // information and will actually create the account.
        ProtocolProviderService protocolProvider
            = super.installAccount(providerFactory, userName, passwd);

        // Check if a default LDAP account should be created for this account.
        String createLdapString = CustomJabberAccRegWizzActivator.getResources()
            .getSettingsString("plugin.customjabberwizz.CREATE_LDAP_ACCOUNT");

        if (protocolProvider != null
            && createLdapString != null
            && new Boolean(createLdapString).booleanValue())
        {
            LdapService ldapService
                = CustomJabberAccRegWizzActivator.getLdapService();

            LdapDirectorySet serverSet =
                CustomJabberAccRegWizzActivator.getLdapService().getServerSet();

            String ldapName = userName + " ("
                + CustomJabberAccRegWizzActivator.getResources()
                    .getSettingsString("plugin.customjabberwizz.LDAP_NAME")
                + ")";

            LdapDirectory server
                = serverSet.getServerWithName(ldapName);

            // If the ldap server exists, we'll reinit it.
            if (server != null)
            {
                serverSet.removeServerWithName(server.getSettings().getName());
                new RefreshContactSourceThread(server, null).start();
            }

            server = ldapService.getFactory().createServer(createLdapSettings(
                        ldapService, ldapName, userName, passwd));

            new RefreshContactSourceThread(null, server).start();
            serverSet.addServer(server);
        }

        return protocolProvider;
    }

    /**
     * Adds a default STUN/TURN server account with the same credentials as the
     * custom jabber account.
     *
     * @param reg the <tt>JabberAccountRegistration</tt> containing all account
     * related information
     */
    private void addDefaultStunServerAccount(JabberAccountRegistration reg)
    {
        reg.addStunServer(new StunServerDescriptor(
            CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.DEFAULT_STUN_SERVER_ADDRESS"),
            new Integer(CustomJabberAccRegWizzActivator.getResources()
                .getSettingsString(
                    "plugin.customjabberwizz.DEFAULT_STUN_SERVER_PORT"))
                    .intValue(),
            new Boolean(CustomJabberAccRegWizzActivator.getResources()
                .getSettingsString(
                    "plugin.customjabberwizz.DEFAULT_STUN_SUPPORT_TURN"))
                    .booleanValue(),
            reg.getUserID(),
            reg.getPassword()));
    }

    /**
     * Creates the LDAP settings.
     *
     * @param ldapService the LDAP service
     * @param ldapName the ldap name
     * @param userName the user name
     * @param passwd the password
     * @return the created ldap settings
     */
    private LdapDirectorySettings createLdapSettings(
                                            LdapService ldapService,
                                            String ldapName,
                                            String userName,
                                            String passwd)
    {
        LdapDirectorySettings ldapSettings
            = ldapService.getFactory().createServerSettings();

        ldapSettings.setName(ldapName);
        ldapSettings.setHostname(
            CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.DEFAULT_LDAP_SERVER"));

        String useSSLString
            = CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.DEFAULT_LDAP_USE_SSL");

        if (useSSLString !=null && useSSLString.equals("true"))
            ldapSettings.setEncryption(Encryption.SSL);
        else
            ldapSettings.setEncryption(Encryption.CLEAR);

        String ldapAuthString
            = CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.DEFAULT_LDAP_AUTH");

        if (ldapAuthString !=null && ldapAuthString.equals("true"))
            ldapSettings.setAuth(Auth.SIMPLE);
        else
            ldapSettings.setAuth(Auth.NONE);

        String ldapPortString
            = CustomJabberAccRegWizzActivator.getResources().getSettingsString(
                "plugin.customjabberwizz.DEFAULT_LDAP_PORT");

        ldapSettings.setPort(new Integer(ldapPortString).intValue());

        String baseDN = CustomJabberAccRegWizzActivator.getResources()
            .getSettingsString("plugin.customjabberwizz.DEFAULT_LDAP_BASE");

        ldapSettings.setBindDN("uid="
            + JabberAccountRegistrationForm.getUserFromUserName(userName)
            + "," + baseDN);

        ldapSettings.setPassword(passwd);

        ldapSettings.setBaseDN(baseDN);

        ldapSettings.setScope(Scope.SUB);

        String mailAttr = CustomJabberAccRegWizzActivator.getResources()
            .getSettingsString("plugin.customjabberwizz.DEFAULT_LDAP_MAIL_ATTR");
        String mailSuffix = CustomJabberAccRegWizzActivator.getResources()
            .getSettingsString("plugin.customjabberwizz.DEFAULT_USER_SUFFIX");
        if(mailAttr != null && mailAttr.length() > 0)
        {
            ldapSettings.setMailSearchFields(stringToList(mailAttr));
        }

        if(mailSuffix != null && mailSuffix.length() > 0)
        {
            mailSuffix = "@" + mailSuffix;
            ldapSettings.setMailSuffix(mailSuffix);
        }

        return ldapSettings;
    }

    /**
     * Thread that will perform refresh of contact sources.
     */
    public static class RefreshContactSourceThread
        extends Thread
    {
        /**
         * LDAP directory to remove.
         */
        private LdapDirectory oldLdap = null;

        /**
         * LDAP directory to add.
         */
        private LdapDirectory newLdap = null;

        /**
         * Constructor.
         *
         * @param oldLdap LDAP directory to remove
         * @param newLdap LDAP directory to add.
         */
        RefreshContactSourceThread(LdapDirectory oldLdap, LdapDirectory newLdap)
        {
            this.oldLdap = oldLdap;
            this.newLdap = newLdap;
        }

        /**
         * Thread entry point.
         */
        public void run()
        {
            LdapService service
                = CustomJabberAccRegWizzActivator.getLdapService();

            if(oldLdap != null)
            {
                service.removeContactSource(oldLdap);
            }

            if(newLdap != null)
            {
                service.createContactSource(newLdap);
            }
        }
    }

    /**
     * Merge String elements separated by space into a List.
     *
     * @param attrs <tt>String</tt>
     * @return list of <tt>String</tt>
     */
    public static List<String> stringToList(String attrs)
    {
        StringTokenizer token = new StringTokenizer(attrs, " ");
        List<String> lst = new ArrayList<String>();

        while(token.hasMoreTokens())
        {
            lst.add(token.nextToken());
        }

        return lst;
    }

    /**
     * Indicates if a wizard is hidden. This may be used if we don't want that
     * a wizard appears in the list of available networks.
     *
     * @return <tt>true</tt> to indicate that a wizard is hidden, <tt>false</tt>
     * otherwise
     */
    public boolean isHidden()
    {
        return false;
    }
}
